# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2024.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

# The original implementation of Qiskit's `parallel_map` in our commit c9c4ed52 was substantially
# derived from QuTiP's (https://github.com/qutip/qutip) in `qutip/parallel.py` at their commit
# f22d3cb7.  It has subsequently been significantly rewritten.
#
# The original implementation was used under these licence terms:
#
#    Copyright (c) 2011 and later, Paul D. Nation and Robert J. Johansson.
#    All rights reserved.
#
#    Redistribution and use in source and binary forms, with or without
#    modification, are permitted provided that the following conditions are
#    met:
#
#    1. Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#
#    2. Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#
#    3. Neither the name of the QuTiP: Quantum Toolbox in Python nor the names
#       of its contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#
#    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
#    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
#    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
#    PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
#    HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#    SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#    LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
#    DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
#    THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#    OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
###############################################################################

"""
Routines for running Python functions in parallel using process pools
from the multiprocessing library.
"""

from __future__ import annotations

import contextlib
import functools
import multiprocessing
import os
import platform
import sys
import warnings
from concurrent.futures import ProcessPoolExecutor

from qiskit import user_config


CONFIG = user_config.get_config()


def _task_wrapper(param):
    (task, value, task_args, task_kwargs) = param
    return task(value, *task_args, **task_kwargs)


def _physical_cpus_assuming_twofold_smt():
    if (sched_getaffinity := getattr(os, "sched_getaffinity", None)) is not None:
        # It is callable, just pylint doesn't recognise it as `os.sched_getaffinity` because of the
        # `getattr`.
        # pylint: disable=not-callable
        num_cpus = len(sched_getaffinity(0))
    else:
        num_cpus = os.cpu_count() or 1
    return (num_cpus // 2) or 1


def _parallel_default():
    # We default to False on `spawn`-based multiprocessing implementations, True on everything else.
    if (set_start_method := multiprocessing.get_start_method(allow_none=True)) is None:
        # The method hasn't been explicitly set, but it would be badly behaved of us to set it for
        # the user, so handle platform defaults.
        return sys.platform not in ("darwin", "win32")
    return set_start_method in ("fork", "forkserver")


@functools.cache
def default_num_processes() -> int:
    """Get the number of processes that a multiprocessing parallel call will use by default.

    Such functions typically also accept a ``num_processes`` keyword argument that will supersede
    the value returned from this function.

    In order of priority (highest to lowest), the return value will be:

    1. The ``QISKIT_NUM_PROCS`` environment variable, if set.
    2. The ``num_processes`` key of the Qiskit user configuration file, if set.
    3. Half of the logical CPUs available to this process, if this can be determined.  This is a
       proxy for the number of physical CPUs, assuming two-fold simultaneous multithreading (SMT);
       empirically, multiprocessing performance of Qiskit seems to be worse when attempting to use
       SMT cores.
    4. 1, if all else fails.

    If a user-configured value is set to a number less than 1, it is treated as if it were 1.
    """
    # Ignore both `None` (unset) and explicit set to empty string.
    if env_num_processes := os.getenv("QISKIT_NUM_PROCS"):
        try:
            env_num_processes = int(env_num_processes)
        except ValueError:
            # Invalid: fall back to other methods.
            warnings.warn(
                "failed to interpret environment 'QISKIT_NUM_PROCS' as a number:"
                f" '{env_num_processes}'"
            )
        else:
            return env_num_processes if env_num_processes > 0 else 1
    if (user_num_processes := CONFIG.get("num_processes", None)) is not None:
        return user_num_processes if user_num_processes > 0 else 1
    return _physical_cpus_assuming_twofold_smt()


def local_hardware_info():
    """Basic hardware information about the local machine.

    Attempts to estimate the number of physical CPUs in the machine, even when hyperthreading is
    turned on. CPU count defaults to 1 when true count can't be determined.

    Returns:
        dict: The hardware information.
    """
    return {
        "python_compiler": platform.python_compiler(),
        "python_build": ", ".join(platform.python_build()),
        "python_version": platform.python_version(),
        "os": platform.system(),
        "cpus": _physical_cpus_assuming_twofold_smt(),
    }


def is_main_process() -> bool:
    """Checks whether the current process is the main one.

    Since Python 3.8, this is identical to the standard Python way of calculating this::

        >>> import multiprocessing
        >>> multiprocessing.parent_process() is None

    This function is left for backwards compatibility, but there is little reason not to use the
    built-in tooling of Python.
    """
    return multiprocessing.parent_process() is None


_PARALLEL_OVERRIDE = None
_PARALLEL_IGNORE_USER_SETTINGS = False
_IN_PARALLEL_ALLOW_PARALLELISM = "FALSE"
_IN_PARALLEL_FORBID_PARALLELISM = "TRUE"


@functools.cache
def should_run_in_parallel(num_processes: int | None = None) -> bool:
    """Decide whether a multiprocessing function should spawn subprocesses for parallelization.

    In particular, this is how :func:`parallel_map` decides whether to use multiprocessing or not.
    The ``num_processes`` argument alone does not enforce parallelism; by default, Qiskit will only
    use process-based parallelism when a ``fork``-like process spawning start method is in effect.
    You can override this decision either by setting the :mod:`multiprocessing` start method you
    use, setting the ``QISKIT_PARALLEL`` environment variable to ``"TRUE"``, or setting
    ``parallel = true`` in your user settings file.

    This function includes two context managers that can be used to temporarily modify the return
    value of this function:

    .. autofunction:: qiskit.utils::should_run_in_parallel.override
    .. autofunction:: qiskit.utils::should_run_in_parallel.ignore_user_settings

    Args:
        num_processes: the maximum number of processes requested for use (``None`` implies the
            default).

    Examples:
        Temporarily override the configured settings to disable parallelism::

            >>> with should_run_in_parallel.override(True):
            ...     assert should_run_in_parallel(8)
            >>> with should_run_in_parallel.override(False):
            ...     assert not should_run_in_parallel(8)
    """
    # It's a configuration function with many simple choices - it'd be less clean to return late.
    # pylint: disable=too-many-return-statements
    num_processes = default_num_processes() if num_processes is None else num_processes
    if num_processes < 2:
        # There's no resources to parallelise over.
        return False
    if (
        os.getenv("QISKIT_IN_PARALLEL", _IN_PARALLEL_ALLOW_PARALLELISM)
        != _IN_PARALLEL_ALLOW_PARALLELISM
    ):
        # This isn't a user-set variable; we set this to talk to our own child processes.
        return False
    if _PARALLEL_OVERRIDE is not None:
        return _PARALLEL_OVERRIDE
    if _PARALLEL_IGNORE_USER_SETTINGS:
        return _parallel_default()
    if (env_qiskit_parallel := os.getenv("QISKIT_PARALLEL")) is not None:
        return env_qiskit_parallel.lower() == "true"
    if (user_qiskit_parallel := CONFIG.get("parallel_enabled", None)) is not None:
        return user_qiskit_parallel
    # Otherwise, fallback to the default.
    return _parallel_default()


@contextlib.contextmanager
def _parallel_ignore_user_settings():
    """A context manager within which :func:`should_run_in_parallel` will ignore environmental
    configuration variables.

    In particular, the ``QISKIT_PARALLEL`` environment variable and the user-configuration file are
    ignored within this context."""
    # The way around this would be to encapsulate `should_run_in_parallel` into a class, but since
    # it's a singleton, it ends up being functionally no different to a global anyway.
    global _PARALLEL_IGNORE_USER_SETTINGS  # pylint: disable=global-statement

    should_run_in_parallel.cache_clear()
    previous, _PARALLEL_IGNORE_USER_SETTINGS = _PARALLEL_IGNORE_USER_SETTINGS, True
    try:
        yield
    finally:
        _PARALLEL_IGNORE_USER_SETTINGS = previous
        should_run_in_parallel.cache_clear()


@contextlib.contextmanager
def _parallel_override(value: bool):
    """A context manager within which :func:`should_run_in_parallel` will return the given
    ``value``.

    This is not a *complete* override; Qiskit will never attempt to parallelize if only a single
    process is available, and will not allow process-based parallelism at a depth greater than 1."""
    # The way around this would be to encapsulate `should_run_in_parallel` into a class, but since
    # it's a singleton, it ends up being functionally no different to a global anyway.
    global _PARALLEL_OVERRIDE  # pylint: disable=global-statement

    should_run_in_parallel.cache_clear()
    previous, _PARALLEL_OVERRIDE = _PARALLEL_OVERRIDE, value
    try:
        yield
    finally:
        _PARALLEL_OVERRIDE = previous
        should_run_in_parallel.cache_clear()


should_run_in_parallel.ignore_user_settings = _parallel_ignore_user_settings
should_run_in_parallel.override = _parallel_override


def parallel_map(task, values, task_args=(), task_kwargs=None, num_processes=None):
    """
    Parallel execution of a mapping of `values` to the function `task`. This
    is functionally equivalent to::

        result = [task(value, *task_args, **task_kwargs) for value in values]

    This will parallelise the results if the number of ``values`` is greater than one and
    :func:`should_run_in_parallel` returns ``True``.  If not, it will run in serial.

    Args:
        task (func): Function that is to be called for each value in ``values``.
        values (array_like): List or array of values for which the ``task`` function is to be
            evaluated.
        task_args (list): Optional additional arguments to the ``task`` function.
        task_kwargs (dict): Optional additional keyword argument to the ``task`` function.
        num_processes (int): Number of processes to spawn.  If not given, the return value of
            :func:`default_num_processes` is used.

    Returns:
        result: The result list contains the value of ``task(value, *task_args, **task_kwargs)`` for
        each value in ``values``.

    Examples:

        .. plot::
           :include-source:
           :nofigs:

            import time
            from qiskit.utils import parallel_map
            def func(_):
                    time.sleep(0.1)
                    return 0
            parallel_map(func, list(range(10)));
    """
    task_kwargs = {} if task_kwargs is None else task_kwargs
    if num_processes is None:
        num_processes = default_num_processes()
    if len(values) < 2 or not should_run_in_parallel(num_processes):
        return [task(value, *task_args, **task_kwargs) for value in values]
    work_items = ((task, value, task_args, task_kwargs) for value in values)

    # This isn't a user-set variable; we set this to talk to our own child processes.
    previous_in_parallel = os.getenv("QISKIT_IN_PARALLEL", _IN_PARALLEL_ALLOW_PARALLELISM)
    os.environ["QISKIT_IN_PARALLEL"] = _IN_PARALLEL_FORBID_PARALLELISM
    try:
        with ProcessPoolExecutor(max_workers=num_processes) as executor:
            return list(executor.map(_task_wrapper, work_items))
    finally:
        os.environ["QISKIT_IN_PARALLEL"] = previous_in_parallel



# This code is part of Qiskit.
#
# (C) Copyright IBM 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Global Phase Gate"""

from typing import Optional

import numpy

from qiskit.circuit.gate import Gate
from qiskit.circuit.parameterexpression import ParameterValueType
from qiskit._accelerate.circuit import StandardGate


class GlobalPhaseGate(Gate):
    r"""The global phase gate (:math:`e^{i\theta}`).

    Can be applied to a :class:`~qiskit.circuit.QuantumCircuit`

    **Mathematical Representation:**

    .. math::
        \text{GlobalPhaseGate}\ =
            \begin{pmatrix}
                e^{i\theta}
            \end{pmatrix}
    """

    _standard_gate = StandardGate.GlobalPhaseGate

    def __init__(self, phase: ParameterValueType, label: Optional[str] = None):
        """
        Args:
            phase: The value of phase it takes.
            label: An optional label for the gate.
        """
        super().__init__("global_phase", 0, [phase], label=label)

    def _define(self):
        # pylint: disable=cyclic-import
        from qiskit.circuit import QuantumCircuit, QuantumRegister

        q = QuantumRegister(0, "q")
        qc = QuantumCircuit(q, name=self.name, global_phase=self.params[0])

        self.definition = qc

    def inverse(self, annotated: bool = False):
        r"""Return inverse GlobalPhaseGate gate.

        :math:`\text{GlobalPhaseGate}(\lambda)^{\dagger} = \text{GlobalPhaseGate}(-\lambda)`

        Args:
            annotated: when set to ``True``, this is typically used to return an
                :class:`.AnnotatedOperation` with an inverse modifier set instead of a concrete
                :class:`.Gate`. However, for this class this argument is ignored as the inverse
                is always another :class:`.GlobalPhaseGate` with an inverted
                parameter value.

        Returns:
            GlobalPhaseGate: inverse gate.
        """
        return GlobalPhaseGate(-self.params[0])

    def __array__(self, dtype=None, copy=None):
        """Return a numpy.array for the global_phase gate."""
        if copy is False:
            raise ValueError("unable to avoid copy while creating an array as requested")
        theta = self.params[0]
        return numpy.array([[numpy.exp(1j * theta)]], dtype=dtype or complex)

    def __eq__(self, other):
        if isinstance(other, GlobalPhaseGate):
            return self._compare_parameters(other)
        return False



# This code is part of Qiskit.
#
# (C) Copyright IBM 2023, 2024.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""
Statevector Sampler V2 class
"""

from __future__ import annotations

import warnings
from dataclasses import dataclass
from typing import Iterable

import numpy as np
from numpy.typing import NDArray

from qiskit import ClassicalRegister, QiskitError, QuantumCircuit
from qiskit.quantum_info import Statevector

from .base import BaseSamplerV2
from .base.validation_v1 import _has_measure
from .containers import (
    BitArray,
    DataBin,
    PrimitiveResult,
    SamplerPubResult,
    SamplerPubLike,
)
from .containers.sampler_pub import SamplerPub
from .containers.bit_array import _min_num_bytes
from .primitive_job import PrimitiveJob
from .utils import bound_circuit_to_instruction


@dataclass
class _MeasureInfo:
    creg_name: str
    num_bits: int
    num_bytes: int
    qreg_indices: list[int]


class StatevectorSampler(BaseSamplerV2):
    """
    Simple implementation of :class:`BaseSamplerV2` using full state vector simulation.

    This class is implemented via :class:`~.Statevector` which turns provided circuits into
    pure state vectors, and is therefore incompatible with mid-circuit measurements (although
    other implementations may be).

    As seen in the example below, this sampler supports providing arrays of parameter value sets to
    bind against a single circuit.

    Each tuple of ``(circuit, <optional> parameter values, <optional> shots)``, called a sampler
    primitive unified bloc (PUB), produces its own array-valued result. The :meth:`~run` method can
    be given many pubs at once.

    .. plot::
       :include-source:
       :nofigs:

        from qiskit.circuit import (
            Parameter, QuantumCircuit, ClassicalRegister, QuantumRegister
        )
        from qiskit.primitives import StatevectorSampler

        import matplotlib.pyplot as plt
        import numpy as np

        # Define our circuit registers, including classical registers
        # called 'alpha' and 'beta'.
        qreg = QuantumRegister(3)
        alpha = ClassicalRegister(2, "alpha")
        beta = ClassicalRegister(1, "beta")

        # Define a quantum circuit with two parameters.
        circuit = QuantumCircuit(qreg, alpha, beta)
        circuit.h(0)
        circuit.cx(0, 1)
        circuit.cx(1, 2)
        circuit.ry(Parameter("a"), 0)
        circuit.rz(Parameter("b"), 0)
        circuit.cx(1, 2)
        circuit.cx(0, 1)
        circuit.h(0)
        circuit.measure([0, 1], alpha)
        circuit.measure([2], beta)

        # Define a sweep over parameter values, where the second axis is over.
        # the two parameters in the circuit.
        params = np.vstack([
            np.linspace(-np.pi, np.pi, 100),
            np.linspace(-4 * np.pi, 4 * np.pi, 100)
        ]).T

        # Instantiate a new statevector simulation based sampler object.
        sampler = StatevectorSampler()

        # Start a job that will return shots for all 100 parameter value sets.
        pub = (circuit, params)
        job = sampler.run([pub], shots=256)

        # Extract the result for the 0th pub (this example only has one pub).
        result = job.result()[0]

        # There is one BitArray object for each ClassicalRegister in the
        # circuit. Here, we can see that the BitArray for alpha contains data
        # for all 100 sweep points, and that it is indeed storing data for 2
        # bits over 256 shots.
        assert result.data.alpha.shape == (100,)
        assert result.data.alpha.num_bits == 2
        assert result.data.alpha.num_shots == 256

        # We can work directly with a binary array in performant applications.
        raw = result.data.alpha.array

        # For small registers where it is anticipated to have many counts
        # associated with the same bitstrings, we can turn the data from,
        # for example, the 22nd sweep index into a dictionary of counts.
        counts = result.data.alpha.get_counts(22)

        # Or, convert into a list of bitstrings that preserve shot order.
        bitstrings = result.data.alpha.get_bitstrings(22)
        print(bitstrings)

    """

    def __init__(self, *, default_shots: int = 1024, seed: np.random.Generator | int | None = None):
        """
        Args:
            default_shots: The default shots for the sampler if not specified during run.
            seed: The seed or Generator object for random number generation.
                If None, a random seeded default RNG will be used.
        """
        self._default_shots = default_shots
        self._seed = seed

    @property
    def default_shots(self) -> int:
        """Return the default shots"""
        return self._default_shots

    @property
    def seed(self) -> np.random.Generator | int | None:
        """Return the seed or Generator object for random number generation."""
        return self._seed

    def run(
        self, pubs: Iterable[SamplerPubLike], *, shots: int | None = None
    ) -> PrimitiveJob[PrimitiveResult[SamplerPubResult]]:
        if shots is None:
            shots = self._default_shots
        coerced_pubs = [SamplerPub.coerce(pub, shots) for pub in pubs]
        if any(len(pub.circuit.cregs) == 0 for pub in coerced_pubs):
            warnings.warn(
                "One of your circuits has no output classical registers and so the result "
                "will be empty. Did you mean to add measurement instructions?",
                UserWarning,
            )

        job = PrimitiveJob(self._run, coerced_pubs)
        job._submit()
        return job

    def _run(self, pubs: Iterable[SamplerPub]) -> PrimitiveResult[SamplerPubResult]:
        results = [self._run_pub(pub) for pub in pubs]
        return PrimitiveResult(results, metadata={"version": 2})

    def _run_pub(self, pub: SamplerPub) -> SamplerPubResult:
        circuit, qargs, meas_info = _preprocess_circuit(pub.circuit)
        bound_circuits = pub.parameter_values.bind_all(circuit)
        arrays = {
            item.creg_name: np.zeros(
                bound_circuits.shape + (pub.shots, item.num_bytes), dtype=np.uint8
            )
            for item in meas_info
        }
        for index, bound_circuit in np.ndenumerate(bound_circuits):
            final_state = Statevector(bound_circuit_to_instruction(bound_circuit))
            final_state.seed(self._seed)
            if qargs:
                samples = final_state.sample_memory(shots=pub.shots, qargs=qargs)
            else:
                samples = [""] * pub.shots
            samples_array = np.array([np.fromiter(sample, dtype=np.uint8) for sample in samples])
            for item in meas_info:
                ary = _samples_to_packed_array(samples_array, item.num_bits, item.qreg_indices)
                arrays[item.creg_name][index] = ary

        meas = {
            item.creg_name: BitArray(arrays[item.creg_name], item.num_bits) for item in meas_info
        }
        return SamplerPubResult(
            DataBin(**meas, shape=pub.shape),
            metadata={"shots": pub.shots, "circuit_metadata": pub.circuit.metadata},
        )


def _preprocess_circuit(circuit: QuantumCircuit):
    num_bits_dict = {creg.name: creg.size for creg in circuit.cregs}
    mapping = _final_measurement_mapping(circuit)
    qargs = sorted(set(mapping.values()))
    qargs_index = {v: k for k, v in enumerate(qargs)}
    circuit = circuit.remove_final_measurements(inplace=False)
    if _has_control_flow(circuit):
        raise QiskitError("StatevectorSampler cannot handle ControlFlowOp")
    if _has_measure(circuit):
        raise QiskitError("StatevectorSampler cannot handle mid-circuit measurements")
    # num_qubits is used as sentinel to fill 0 in _samples_to_packed_array
    sentinel = len(qargs)
    indices = {key: [sentinel] * val for key, val in num_bits_dict.items()}
    for key, qreg in mapping.items():
        creg, ind = key
        indices[creg.name][ind] = qargs_index[qreg]
    meas_info = [
        _MeasureInfo(
            creg_name=name,
            num_bits=num_bits,
            num_bytes=_min_num_bytes(num_bits),
            qreg_indices=indices[name],
        )
        for name, num_bits in num_bits_dict.items()
    ]
    return circuit, qargs, meas_info


def _samples_to_packed_array(
    samples: NDArray[np.uint8], num_bits: int, indices: list[int]
) -> NDArray[np.uint8]:
    # samples of `Statevector.sample_memory` will be in the order of
    # qubit_last, ..., qubit_1, qubit_0.
    # reverse the sample order into qubit_0, qubit_1, ..., qubit_last and
    # pad 0 in the rightmost to be used for the sentinel introduced by _preprocess_circuit.
    ary = np.pad(samples[:, ::-1], ((0, 0), (0, 1)), constant_values=0)
    # place samples in the order of clbit_last, ..., clbit_1, clbit_0
    ary = ary[:, indices[::-1]]
    # pad 0 in the left to align the number to be mod 8
    # since np.packbits(bitorder='big') pads 0 to the right.
    pad_size = -num_bits % 8
    ary = np.pad(ary, ((0, 0), (pad_size, 0)), constant_values=0)
    # pack bits in big endian order
    ary = np.packbits(ary, axis=-1)
    return ary


def _final_measurement_mapping(circuit: QuantumCircuit) -> dict[tuple[ClassicalRegister, int], int]:
    """Return the final measurement mapping for the circuit.

    Parameters:
        circuit: Input quantum circuit.

    Returns:
        Mapping of classical bits to qubits for final measurements.
    """
    active_qubits = set(range(circuit.num_qubits))
    active_cbits = set(range(circuit.num_clbits))

    # Find final measurements starting in back
    mapping = {}
    for item in circuit[::-1]:
        if item.operation.name == "measure":
            loc = circuit.find_bit(item.clbits[0])
            cbit = loc.index
            qbit = circuit.find_bit(item.qubits[0]).index
            if cbit in active_cbits and qbit in active_qubits:
                for creg in loc.registers:
                    mapping[creg] = qbit
                active_cbits.remove(cbit)
        elif item.operation.name not in ["barrier", "delay"]:
            for qq in item.qubits:
                _temp_qubit = circuit.find_bit(qq).index
                if _temp_qubit in active_qubits:
                    active_qubits.remove(_temp_qubit)

        if not active_cbits or not active_qubits:
            break

    return mapping


def _has_control_flow(circuit: QuantumCircuit) -> bool:
    return any(instruction.is_control_flow() for instruction in circuit)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2019.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

# pylint: disable=too-many-return-statements

"""
QPY Type keys for several namespace.
"""

import uuid
from abc import abstractmethod
from enum import Enum, IntEnum

import numpy as np

from qiskit.circuit import (
    Gate,
    Instruction,
    QuantumCircuit,
    ControlledGate,
    CASE_DEFAULT,
    Clbit,
    ClassicalRegister,
    Duration,
)
from qiskit.circuit.annotated_operation import AnnotatedOperation, Modifier
from qiskit.circuit.classical import expr, types
from qiskit.circuit.library import PauliEvolutionGate
from qiskit.circuit.parameter import Parameter
from qiskit.circuit.parameterexpression import ParameterExpression
from qiskit.circuit.parametervector import ParameterVectorElement
from qiskit.qpy import exceptions


class TypeKeyBase(bytes, Enum):
    """Abstract baseclass for type key Enums."""

    @classmethod
    @abstractmethod
    def assign(cls, obj):
        """Assign type key to given object.

        Args:
            obj (any): Arbitrary object to evaluate.

        Returns:
            TypeKey: Corresponding key object.
        """
        pass

    @classmethod
    @abstractmethod
    def retrieve(cls, type_key):
        """Get a class from given type key.

        Args:
            type_key (bytes): Object type key.

        Returns:
            any: Corresponding class.
        """
        pass


class Value(TypeKeyBase):
    """Type key enum for value object."""

    INTEGER = b"i"
    FLOAT = b"f"
    COMPLEX = b"c"
    CASE_DEFAULT = b"d"
    REGISTER = b"R"
    NUMPY_OBJ = b"n"
    PARAMETER = b"p"
    PARAMETER_VECTOR = b"v"
    PARAMETER_EXPRESSION = b"e"
    STRING = b"s"
    NULL = b"z"
    EXPRESSION = b"x"
    MODIFIER = b"m"

    @classmethod
    def assign(cls, obj):
        if isinstance(obj, int):
            return cls.INTEGER
        if isinstance(obj, float):
            return cls.FLOAT
        if isinstance(obj, complex):
            return cls.COMPLEX
        if isinstance(obj, (np.integer, np.floating, np.complexfloating, np.ndarray)):
            return cls.NUMPY_OBJ
        if isinstance(obj, ParameterVectorElement):
            return cls.PARAMETER_VECTOR
        if isinstance(obj, Parameter):
            return cls.PARAMETER
        if isinstance(obj, ParameterExpression):
            return cls.PARAMETER_EXPRESSION
        if isinstance(obj, str):
            return cls.STRING
        if isinstance(obj, (Clbit, ClassicalRegister)):
            return cls.REGISTER
        if obj is None:
            return cls.NULL
        if obj is CASE_DEFAULT:
            return cls.CASE_DEFAULT
        if isinstance(obj, expr.Expr):
            return cls.EXPRESSION
        if isinstance(obj, Modifier):
            return cls.MODIFIER

        raise exceptions.QpyError(
            f"Object type '{type(obj)}' is not supported in {cls.__name__} namespace."
        )

    @classmethod
    def retrieve(cls, type_key):
        raise NotImplementedError


class Condition(IntEnum):
    """Type keys for the ``conditional_key`` field of the INSTRUCTION struct."""

    # This class is deliberately raw integers and not in terms of ASCII characters for backwards
    # compatibility in the form as an old Boolean value was expanded; `NONE` and `TWO_TUPLE` must
    # have the enumeration values 0 and 1.

    NONE = 0
    TWO_TUPLE = 1
    EXPRESSION = 2


class Container(TypeKeyBase):
    """Type key enum for container-like object."""

    RANGE = b"r"
    TUPLE = b"t"

    @classmethod
    def assign(cls, obj):
        if isinstance(obj, range):
            return cls.RANGE
        if isinstance(obj, tuple):
            return cls.TUPLE

        raise exceptions.QpyError(
            f"Object type '{type(obj)}' is not supported in {cls.__name__} namespace."
        )

    @classmethod
    def retrieve(cls, type_key):
        raise NotImplementedError


class CircuitInstruction(TypeKeyBase):
    """Type key enum for circuit instruction object."""

    INSTRUCTION = b"i"
    GATE = b"g"
    PAULI_EVOL_GATE = b"p"
    CONTROLLED_GATE = b"c"
    ANNOTATED_OPERATION = b"a"

    @classmethod
    def assign(cls, obj):
        if isinstance(obj, PauliEvolutionGate):
            return cls.PAULI_EVOL_GATE
        if isinstance(obj, ControlledGate):
            return cls.CONTROLLED_GATE
        if isinstance(obj, AnnotatedOperation):
            return cls.ANNOTATED_OPERATION
        if isinstance(obj, Gate):
            return cls.GATE
        if isinstance(obj, Instruction):
            return cls.INSTRUCTION

        raise exceptions.QpyError(
            f"Object type '{type(obj)}' is not supported in {cls.__name__} namespace."
        )

    @classmethod
    def retrieve(cls, type_key):
        raise NotImplementedError


class ScheduleOperand(TypeKeyBase):
    """Type key enum for schedule instruction operand object.

    Note: This class is kept post pulse-removal to allow reading of
    legacy payloads containing pulse gates without breaking the entire
    load flow.
    """

    WAVEFORM = b"w"
    SYMBOLIC_PULSE = b"s"
    CHANNEL = b"c"
    KERNEL = b"k"
    DISCRIMINATOR = b"d"

    # We need to have own string type definition for operands of schedule instruction.
    # Note that string type is already defined in the Value namespace,
    # but its key "s" conflicts with the SYMBOLIC_PULSE in the ScheduleOperand namespace.
    # New in QPY version 7.
    OPERAND_STR = b"o"

    @classmethod
    def assign(cls, _):
        raise NotImplementedError

    @classmethod
    def retrieve(cls, _):
        raise NotImplementedError


class Program(TypeKeyBase):
    """Type key enum for program that QPY supports."""

    CIRCUIT = b"q"
    # This is left for backward compatibility, for identifying payloads of type `ScheduleBlock`
    # and raising accordingly. `ScheduleBlock` support has been removed in Qiskit 2.0 as part
    # of the pulse package removal in that version.
    SCHEDULE_BLOCK = b"s"

    @classmethod
    def assign(cls, obj):
        if isinstance(obj, QuantumCircuit):
            return cls.CIRCUIT

        raise exceptions.QpyError(
            f"Object type '{type(obj)}' is not supported in {cls.__name__} namespace."
        )

    @classmethod
    def retrieve(cls, type_key):
        raise NotImplementedError


class Expression(TypeKeyBase):
    """Type keys for the ``EXPRESSION`` QPY item."""

    VAR = b"x"
    STRETCH = b"s"
    VALUE = b"v"
    CAST = b"c"
    UNARY = b"u"
    BINARY = b"b"
    INDEX = b"i"

    @classmethod
    def assign(cls, obj):
        if (
            isinstance(obj, expr.Expr)
            and (key := getattr(cls, obj.__class__.__name__.upper(), None)) is not None
        ):
            return key
        raise exceptions.QpyError(f"Object '{obj}' is not supported in {cls.__name__} namespace.")

    @classmethod
    def retrieve(cls, type_key):
        raise NotImplementedError


class ExprVarDeclaration(TypeKeyBase):
    """Type keys for the ``EXPR_VAR_DECLARATION`` QPY item."""

    INPUT = b"I"
    CAPTURE = b"C"
    LOCAL = b"L"
    STRETCH_CAPTURE = b"A"
    STRETCH_LOCAL = b"O"

    @classmethod
    def assign(cls, obj):
        raise NotImplementedError

    @classmethod
    def retrieve(cls, type_key):
        raise NotImplementedError


class ExprType(TypeKeyBase):
    """Type keys for the ``EXPR_TYPE`` QPY item."""

    BOOL = b"b"
    UINT = b"u"
    FLOAT = b"f"
    DURATION = b"d"

    @classmethod
    def assign(cls, obj):
        if (
            isinstance(obj, types.Type)
            and (key := getattr(cls, obj.__class__.__name__.upper(), None)) is not None
        ):
            return key
        raise exceptions.QpyError(f"Object '{obj}' is not supported in {cls.__name__} namespace.")

    @classmethod
    def retrieve(cls, type_key):
        raise NotImplementedError


class ExprVar(TypeKeyBase):
    """Type keys for the ``EXPR_VAR`` QPY item."""

    CLBIT = b"C"
    REGISTER = b"R"
    UUID = b"U"

    @classmethod
    def assign(cls, obj):
        if isinstance(obj, uuid.UUID):
            return cls.UUID
        if isinstance(obj, Clbit):
            return cls.CLBIT
        if isinstance(obj, ClassicalRegister):
            return cls.REGISTER
        raise exceptions.QpyError(
            f"Object type '{type(obj)}' is not supported in {cls.__name__} namespace."
        )

    @classmethod
    def retrieve(cls, type_key):
        raise NotImplementedError


class ExprValue(TypeKeyBase):
    """Type keys for the ``EXPR_VALUE`` QPY item."""

    BOOL = b"b"
    INT = b"i"
    FLOAT = b"f"
    DURATION = b"t"

    @classmethod
    def assign(cls, obj):
        if isinstance(obj, bool):
            return cls.BOOL
        if isinstance(obj, int):
            return cls.INT
        if isinstance(obj, float):
            return cls.FLOAT
        if isinstance(obj, Duration):
            return cls.DURATION
        raise exceptions.QpyError(
            f"Object type '{type(obj)}' is not supported in {cls.__name__} namespace."
        )

    @classmethod
    def retrieve(cls, type_key):
        raise NotImplementedError


class CircuitDuration(TypeKeyBase):
    """Type keys for the ``DURATION`` QPY item."""

    DT = b"t"
    NS = b"n"
    US = b"u"
    MS = b"m"
    S = b"s"

    @classmethod
    def assign(cls, obj):
        if isinstance(obj, Duration):
            unit = obj.unit()
            if unit == "dt":
                return cls.DT
            if unit == "ns":
                return cls.NS
            if unit == "us":
                return cls.US
            if unit == "ms":
                return cls.MS
            if unit == "s":
                return cls.S
        raise exceptions.QpyError(
            f"Object type '{type(obj)}' is not supported in {cls.__name__} namespace."
        )

    @classmethod
    def retrieve(cls, type_key):
        raise NotImplementedError


class SymExprEncoding(TypeKeyBase):
    """Type keys for the symbolic encoding field in the file header."""

    SYMPY = b"p"
    SYMENGINE = b"e"

    @classmethod
    def assign(cls, obj):
        if obj:
            return cls.SYMENGINE
        else:
            return cls.SYMPY

    @classmethod
    def retrieve(cls, type_key):
        raise NotImplementedError



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2025.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Implement a integer-weighted sum over a set of qubits."""

from __future__ import annotations

import typing
import numpy as np
from qiskit.circuit.quantumcircuit import QuantumCircuit, QuantumRegister, AncillaRegister
from qiskit.synthesis.multi_controlled import synth_mcx_n_clean_m15

if typing.TYPE_CHECKING:
    from qiskit.circuit.library import WeightedSumGate


def synth_weighted_sum_carry(weighted_sum: WeightedSumGate) -> QuantumCircuit:
    """Synthesize a weighted sum gate, by the number of state qubits and the qubit weights.

    This method is described in Appendix A of [1].

    Reference:

        [1] Stamatopoulos et al. Option Pricing using Quantum Computers (2020)
            `Quantum 4, 291 <https://doi.org/10.22331/q-2020-07-06-291>`__
    """

    num_sum_qubits = weighted_sum.num_sum_qubits
    num_state_qubits = weighted_sum.num_qubits - num_sum_qubits

    num_carry_qubits = num_sum_qubits - 1
    num_control_qubits = int(num_sum_qubits > 2)

    qr_state = QuantumRegister(num_state_qubits, "state")
    qr_sum = QuantumRegister(num_sum_qubits, "sum")
    qr_carry = AncillaRegister(num_carry_qubits, "carry")
    qr_control = AncillaRegister(num_control_qubits, "control")
    circuit = QuantumCircuit(qr_state, qr_sum, qr_carry, qr_control)

    # we use a MCX v-chain decomposition of C3X here, synthesize it once and re-use
    c3x = synth_mcx_n_clean_m15(3)

    # loop over state qubits and corresponding weights
    weights = weighted_sum.params
    for i, weight in enumerate(weights):
        # only act if non-trivial weight
        if np.isclose(weight, 0):
            continue

        # get state control qubit
        q_state = qr_state[i]

        # get bit representation of current weight
        weight_binary = f"{int(weight):b}".rjust(num_sum_qubits, "0")[::-1]

        # loop over bits of current weight and add them to sum and carry registers
        for j, bit in enumerate(weight_binary):
            if bit == "1":
                if num_sum_qubits == 1:
                    circuit.cx(q_state, qr_sum[j])
                elif j == 0:
                    # compute (q_sum[0] + 1) into (q_sum[0], q_carry[0])
                    # - controlled by q_state[i]
                    circuit.ccx(q_state, qr_sum[j], qr_carry[j])
                    circuit.cx(q_state, qr_sum[j])
                elif j == num_sum_qubits - 1:
                    # compute (q_sum[j] + q_carry[j-1] + 1) into (q_sum[j])
                    # - controlled by q_state[i] / last qubit,
                    # no carry needed by construction
                    circuit.cx(q_state, qr_sum[j])
                    circuit.ccx(q_state, qr_carry[j - 1], qr_sum[j])
                else:
                    # compute (q_sum[j] + q_carry[j-1] + 1) into (q_sum[j], q_carry[j])
                    # - controlled by q_state[i]
                    circuit.x(qr_sum[j])
                    circuit.x(qr_carry[j - 1])
                    circuit.compose(
                        c3x,
                        [q_state, qr_sum[j], qr_carry[j - 1], qr_carry[j]] + qr_control[:],
                        inplace=True,
                    )
                    circuit.cx(q_state, qr_carry[j])
                    circuit.x(qr_sum[j])
                    circuit.x(qr_carry[j - 1])
                    circuit.cx(q_state, qr_sum[j])
                    circuit.ccx(q_state, qr_carry[j - 1], qr_sum[j])
            else:
                if num_sum_qubits == 1:
                    pass  # nothing to do, since nothing to add
                elif j == 0:
                    pass  # nothing to do, since nothing to add
                elif j == num_sum_qubits - 1:
                    # compute (q_sum[j] + q_carry[j-1]) into (q_sum[j])
                    # - controlled by q_state[i] / last qubit,
                    # no carry needed by construction
                    circuit.ccx(q_state, qr_carry[j - 1], qr_sum[j])
                else:
                    # compute (q_sum[j] + q_carry[j-1]) into (q_sum[j], q_carry[j])
                    # - controlled by q_state[i]
                    circuit.compose(
                        c3x,
                        [q_state, qr_sum[j], qr_carry[j - 1], qr_carry[j]] + qr_control[:],
                        inplace=True,
                    )
                    circuit.ccx(q_state, qr_carry[j - 1], qr_sum[j])

        # uncompute carry qubits
        for j in reversed(range(len(weight_binary))):
            bit = weight_binary[j]
            if bit == "1":
                if num_sum_qubits == 1:
                    pass
                elif j == 0:
                    circuit.x(qr_sum[j])
                    circuit.ccx(q_state, qr_sum[j], qr_carry[j])
                    circuit.x(qr_sum[j])
                elif j == num_sum_qubits - 1:
                    pass
                else:
                    circuit.x(qr_carry[j - 1])
                    circuit.compose(
                        c3x,
                        [q_state, qr_sum[j], qr_carry[j - 1], qr_carry[j]] + qr_control[:],
                        inplace=True,
                    )
                    circuit.cx(q_state, qr_carry[j])
                    circuit.x(qr_carry[j - 1])
            else:
                if num_sum_qubits == 1:
                    pass
                elif j == 0:
                    pass
                elif j == num_sum_qubits - 1:
                    pass
                else:
                    # compute (q_sum[j] + q_carry[j-1]) into (q_sum[j], q_carry[j])
                    # - controlled by q_state[i]
                    circuit.x(qr_sum[j])
                    circuit.compose(
                        c3x,
                        [q_state, qr_sum[j], qr_carry[j - 1], qr_carry[j]] + qr_control[:],
                        inplace=True,
                    )
                    circuit.x(qr_sum[j])

    return circuit



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

# pylint: disable=invalid-name

"""Test QuantumCircuit.compose()."""

import unittest

import numpy as np

from qiskit.circuit import (
    QuantumRegister,
    ClassicalRegister,
    Clbit,
    QuantumCircuit,
    Qubit,
    Parameter,
    Gate,
    Instruction,
    CASE_DEFAULT,
    SwitchCaseOp,
    CircuitError,
)
from qiskit.circuit.library import HGate, RZGate, CXGate, CCXGate, TwoLocal
from qiskit.circuit.classical import expr, types
from test import QiskitTestCase  # pylint: disable=wrong-import-order


class TestCircuitCompose(QiskitTestCase):
    """Test composition of two circuits."""

    def setUp(self):
        super().setUp()
        qreg1 = QuantumRegister(3, "lqr_1")
        qreg2 = QuantumRegister(2, "lqr_2")
        creg = ClassicalRegister(2, "lcr")

        self.circuit_left = QuantumCircuit(qreg1, qreg2, creg)
        self.circuit_left.h(qreg1[0])
        self.circuit_left.x(qreg1[1])
        self.circuit_left.p(0.1, qreg1[2])
        self.circuit_left.cx(qreg2[0], qreg2[1])

        self.left_qubit0 = qreg1[0]
        self.left_qubit1 = qreg1[1]
        self.left_qubit2 = qreg1[2]
        self.left_qubit3 = qreg2[0]
        self.left_qubit4 = qreg2[1]
        self.left_clbit0 = creg[0]
        self.left_clbit1 = creg[1]
        self.condition = (creg, 3)

    def test_compose_inorder(self):
        """Composing two circuits of the same width, default order.

                      ┌───┐
        lqr_1_0: |0>──┤ H ├───     rqr_0: |0>──■───────
                      ├───┤                    │  ┌───┐
        lqr_1_1: |0>──┤ X ├───     rqr_1: |0>──┼──┤ X ├
                    ┌─┴───┴──┐                 │  ├───┤
        lqr_1_2: |0>┤ P(0.1) ├  +  rqr_2: |0>──┼──┤ Y ├  =
                    └────────┘               ┌─┴─┐└───┘
        lqr_2_0: |0>────■─────     rqr_3: |0>┤ X ├─────
                      ┌─┴─┐                  └───┘┌───┐
        lqr_2_1: |0>──┤ X ├───     rqr_4: |0>─────┤ Z ├
                      └───┘                       └───┘
        lcr_0: 0 ═══════════

        lcr_1: 0 ═══════════


                       ┌───┐
         lqr_1_0: |0>──┤ H ├─────■───────
                       ├───┤     │  ┌───┐
         lqr_1_1: |0>──┤ X ├─────┼──┤ X ├
                     ┌─┴───┴──┐  │  ├───┤
         lqr_1_2: |0>┤ P(0.1) ├──┼──┤ Y ├
                     └────────┘┌─┴─┐└───┘
         lqr_2_0: |0>────■─────┤ X ├─────
                       ┌─┴─┐   └───┘┌───┐
         lqr_2_1: |0>──┤ X ├────────┤ Z ├
                       └───┘        └───┘
         lcr_0: 0 ═══════════════════════

         lcr_1: 0 ═══════════════════════

        """
        qreg = QuantumRegister(5, "rqr")
        circuit_right = QuantumCircuit(qreg)
        circuit_right.cx(qreg[0], qreg[3])
        circuit_right.x(qreg[1])
        circuit_right.y(qreg[2])
        circuit_right.z(qreg[4])

        circuit_expected = self.circuit_left.copy()
        circuit_expected.cx(self.left_qubit0, self.left_qubit3)
        circuit_expected.x(self.left_qubit1)
        circuit_expected.y(self.left_qubit2)
        circuit_expected.z(self.left_qubit4)

        circuit_composed = self.circuit_left.compose(circuit_right, inplace=False)
        self.assertEqual(circuit_composed, circuit_expected)

    def test_compose_inorder_unusual_types(self):
        """Test that composition works in order, using Numpy integer types as well as regular
        integer types.  In general, it should be permissible to use any of the same `QubitSpecifier`
        types (or similar for `Clbit`) that `QuantumCircuit.append` uses."""
        qreg = QuantumRegister(5, "rqr")
        creg = ClassicalRegister(2, "rcr")
        circuit_right = QuantumCircuit(qreg, creg)
        circuit_right.cx(qreg[0], qreg[3])
        circuit_right.x(qreg[1])
        circuit_right.y(qreg[2])
        circuit_right.z(qreg[4])
        circuit_right.measure([0, 1], [0, 1])

        circuit_expected = self.circuit_left.copy()
        circuit_expected.cx(self.left_qubit0, self.left_qubit3)
        circuit_expected.x(self.left_qubit1)
        circuit_expected.y(self.left_qubit2)
        circuit_expected.z(self.left_qubit4)
        circuit_expected.measure(self.left_qubit0, self.left_clbit0)
        circuit_expected.measure(self.left_qubit1, self.left_clbit1)

        circuit_composed = self.circuit_left.compose(circuit_right, np.arange(5), slice(0, 2))
        self.assertEqual(circuit_composed, circuit_expected)

    def test_compose_inorder_inplace(self):
        """Composing two circuits of the same width, default order, inplace.

                       ┌───┐
        lqr_1_0: |0>───┤ H ├───     rqr_0: |0>──■───────
                       ├───┤                    │  ┌───┐
        lqr_1_1: |0>───┤ X ├───     rqr_1: |0>──┼──┤ X ├
                    ┌──┴───┴──┐                 │  ├───┤
        lqr_1_2: |0>┤ U1(0.1) ├  +  rqr_2: |0>──┼──┤ Y ├  =
                    └─────────┘               ┌─┴─┐└───┘
        lqr_2_0: |0>─────■─────     rqr_3: |0>┤ X ├─────
                       ┌─┴─┐                  └───┘┌───┐
        lqr_2_1: |0>───┤ X ├───     rqr_4: |0>─────┤ Z ├
                       └───┘                       └───┘
        lcr_0: 0 ═══════════

        lcr_1: 0 ═══════════


                        ┌───┐
         lqr_1_0: |0>───┤ H ├─────■───────
                        ├───┤     │  ┌───┐
         lqr_1_1: |0>───┤ X ├─────┼──┤ X ├
                     ┌──┴───┴──┐  │  ├───┤
         lqr_1_2: |0>┤ U1(0.1) ├──┼──┤ Y ├
                     └─────────┘┌─┴─┐└───┘
         lqr_2_0: |0>─────■─────┤ X ├─────
                        ┌─┴─┐   └───┘┌───┐
         lqr_2_1: |0>───┤ X ├────────┤ Z ├
                        └───┘        └───┘
         lcr_0: 0 ════════════════════════

         lcr_1: 0 ════════════════════════

        """
        qreg = QuantumRegister(5, "rqr")
        circuit_right = QuantumCircuit(qreg)
        circuit_right.cx(qreg[0], qreg[3])
        circuit_right.x(qreg[1])
        circuit_right.y(qreg[2])
        circuit_right.z(qreg[4])

        circuit_expected = self.circuit_left.copy()
        circuit_expected.cx(self.left_qubit0, self.left_qubit3)
        circuit_expected.x(self.left_qubit1)
        circuit_expected.y(self.left_qubit2)
        circuit_expected.z(self.left_qubit4)

        # inplace
        circuit_left = self.circuit_left.copy()
        circuit_left.compose(circuit_right, inplace=True)
        self.assertEqual(circuit_left, circuit_expected)

    def test_compose_inorder_smaller(self):
        """Composing with a smaller RHS dag, default order.

                       ┌───┐                       ┌─────┐
        lqr_1_0: |0>───┤ H ├───     rqr_0: |0>──■──┤ Tdg ├
                       ├───┤                  ┌─┴─┐└─────┘
        lqr_1_1: |0>───┤ X ├───     rqr_1: |0>┤ X ├───────
                    ┌──┴───┴──┐               └───┘
        lqr_1_2: |0>┤ U1(0.1) ├  +                          =
                    └─────────┘
        lqr_2_0: |0>─────■─────
                       ┌─┴─┐
        lqr_2_1: |0>───┤ X ├───
                       └───┘
        lcr_0: 0 ══════════════

        lcr_1: 0 ══════════════

                        ┌───┐        ┌─────┐
         lqr_1_0: |0>───┤ H ├─────■──┤ Tdg ├
                        ├───┤   ┌─┴─┐└─────┘
         lqr_1_1: |0>───┤ X ├───┤ X ├───────
                     ┌──┴───┴──┐└───┘
         lqr_1_2: |0>┤ U1(0.1) ├────────────
                     └─────────┘
         lqr_2_0: |0>─────■─────────────────
                        ┌─┴─┐
         lqr_2_1: |0>───┤ X ├───────────────
                        └───┘
         lcr_0: 0 ══════════════════════════

         lcr_1: 0 ══════════════════════════

        """
        qreg = QuantumRegister(2, "rqr")

        circuit_right = QuantumCircuit(qreg)
        circuit_right.cx(qreg[0], qreg[1])
        circuit_right.tdg(qreg[0])

        circuit_expected = self.circuit_left.copy()
        circuit_expected.cx(self.left_qubit0, self.left_qubit1)
        circuit_expected.tdg(self.left_qubit0)

        circuit_composed = self.circuit_left.compose(circuit_right)
        self.assertEqual(circuit_composed, circuit_expected)

    def test_compose_permuted(self):
        """Composing two dags of the same width, permuted wires.
                      ┌───┐
        lqr_1_0: |0>──┤ H ├───      rqr_0: |0>──■───────
                      ├───┤                     │  ┌───┐
        lqr_1_1: |0>──┤ X ├───      rqr_1: |0>──┼──┤ X ├
                    ┌─┴───┴──┐                  │  ├───┤
        lqr_1_2: |0>┤ P(0.1) ├      rqr_2: |0>──┼──┤ Y ├
                    └────────┘                ┌─┴─┐└───┘
        lqr_2_0: |0>────■─────  +   rqr_3: |0>┤ X ├─────   =
                      ┌─┴─┐                   └───┘┌───┐
        lqr_2_1: |0>──┤ X ├───      rqr_4: |0>─────┤ Z ├
                      └───┘                        └───┘
        lcr_0: 0 ══════════════

        lcr_1: 0 ══════════════

                      ┌───┐   ┌───┐
        lqr_1_0: |0>──┤ H ├───┤ Z ├
                      ├───┤   ├───┤
        lqr_1_1: |0>──┤ X ├───┤ X ├
                    ┌─┴───┴──┐├───┤
        lqr_1_2: |0>┤ P(0.1) ├┤ Y ├
                    └────────┘└───┘
        lqr_2_0: |0>────■───────■──
                      ┌─┴─┐   ┌─┴─┐
        lqr_2_1: |0>──┤ X ├───┤ X ├
                      └───┘   └───┘
        lcr_0: 0 ══════════════════

        lcr_1: 0 ══════════════════
        """
        qreg = QuantumRegister(5, "rqr")

        circuit_right = QuantumCircuit(qreg)
        circuit_right.cx(qreg[0], qreg[3])
        circuit_right.x(qreg[1])
        circuit_right.y(qreg[2])
        circuit_right.z(qreg[4])

        circuit_expected = self.circuit_left.copy()
        circuit_expected.z(self.left_qubit0)
        circuit_expected.x(self.left_qubit1)
        circuit_expected.y(self.left_qubit2)
        circuit_expected.cx(self.left_qubit3, self.left_qubit4)

        # permuted wiring
        circuit_composed = self.circuit_left.compose(
            circuit_right,
            qubits=[
                self.left_qubit3,
                self.left_qubit1,
                self.left_qubit2,
                self.left_qubit4,
                self.left_qubit0,
            ],
            inplace=False,
        )
        self.assertEqual(circuit_composed, circuit_expected)

    def test_compose_permuted_smaller(self):
        """Composing with a smaller RHS dag, and permuted wires.
        Compose using indices.

                      ┌───┐                       ┌─────┐
        lqr_1_0: |0>──┤ H ├───     rqr_0: |0>──■──┤ Tdg ├
                      ├───┤                  ┌─┴─┐└─────┘
        lqr_1_1: |0>──┤ X ├───     rqr_1: |0>┤ X ├───────
                    ┌─┴───┴──┐               └───┘
        lqr_1_2: |0>┤ P(0.1) ├  +                          =
                    └────────┘
        lqr_2_0: |0>────■─────
                      ┌─┴─┐
        lqr_2_1: |0>──┤ X ├───
                      └───┘
        lcr_0: 0 ═════════════

        lcr_1: 0 ═════════════

                       ┌───┐
         lqr_1_0: |0>──┤ H ├───────────────
                       ├───┤
         lqr_1_1: |0>──┤ X ├───────────────
                     ┌─┴───┴──┐┌───┐
         lqr_1_2: |0>┤ P(0.1) ├┤ X ├───────
                     └────────┘└─┬─┘┌─────┐
         lqr_2_0: |0>────■───────■──┤ Tdg ├
                       ┌─┴─┐        └─────┘
         lqr_2_1: |0>──┤ X ├───────────────
                       └───┘
         lcr_0: 0 ═════════════════════════

         lcr_1: 0 ═════════════════════════
        """
        qreg = QuantumRegister(2, "rqr")
        circuit_right = QuantumCircuit(qreg)
        circuit_right.cx(qreg[0], qreg[1])
        circuit_right.tdg(qreg[0])

        # permuted wiring of subset
        circuit_composed = self.circuit_left.compose(circuit_right, qubits=[3, 2])

        circuit_expected = self.circuit_left.copy()
        circuit_expected.cx(self.left_qubit3, self.left_qubit2)
        circuit_expected.tdg(self.left_qubit3)

        self.assertEqual(circuit_composed, circuit_expected)

    def test_compose_copy(self):
        """Test that `compose` copies instructions where appropriate."""
        base = QuantumCircuit(2, 2)

        # If given a parametric instruction, the instruction should be copied in the output unless
        # specifically set to take ownership.
        parametric = QuantumCircuit(1)
        parametric.rz(Parameter("x"), 0)
        should_copy = base.compose(parametric, qubits=[0])
        self.assertIsNot(should_copy.data[-1].operation, parametric.data[-1].operation)
        self.assertEqual(should_copy.data[-1].operation, parametric.data[-1].operation)
        forbid_copy = base.compose(parametric, qubits=[0], copy=False)
        # For standard gates a fresh copy is returned from the data list each time
        self.assertEqual(forbid_copy.data[-1].operation, parametric.data[-1].operation)

        class Custom(Gate):
            """Custom gate that cannot be decomposed into Rust space."""

            def __init__(self):
                super().__init__("mygate", 1, [])

        conditional = QuantumCircuit(1, 1)
        conditional.append(Custom(), [0], [])
        test = base.compose(conditional, qubits=[0], clbits=[0], copy=False)
        self.assertIs(test.data[-1].operation, conditional.data[-1].operation)

    def test_compose_classical(self):
        """Composing on classical bits.

                      ┌───┐                       ┌─────┐┌─┐
        lqr_1_0: |0>──┤ H ├───     rqr_0: |0>──■──┤ Tdg ├┤M├
                      ├───┤                  ┌─┴─┐└─┬─┬─┘└╥┘
        lqr_1_1: |0>──┤ X ├───     rqr_1: |0>┤ X ├──┤M├───╫─
                    ┌─┴───┴──┐               └───┘  └╥┘   ║
        lqr_1_2: |0>┤ P(0.1) ├  +   rcr_0: 0 ════════╬════╩═  =
                    └────────┘                       ║
        lqr_2_0: |0>────■─────      rcr_1: 0 ════════╩══════
                      ┌─┴─┐
        lqr_2_1: |0>──┤ X ├───
                      └───┘
        lcr_0: 0 ══════════════

        lcr_1: 0 ══════════════

                      ┌───┐
        lqr_1_0: |0>──┤ H ├──────────────────
                      ├───┤        ┌─────┐┌─┐
        lqr_1_1: |0>──┤ X ├─────■──┤ Tdg ├┤M├
                    ┌─┴───┴──┐  │  └─────┘└╥┘
        lqr_1_2: |0>┤ P(0.1) ├──┼──────────╫─
                    └────────┘  │          ║
        lqr_2_0: |0>────■───────┼──────────╫─
                      ┌─┴─┐   ┌─┴─┐  ┌─┐   ║
        lqr_2_1: |0>──┤ X ├───┤ X ├──┤M├───╫─
                      └───┘   └───┘  └╥┘   ║
           lcr_0: 0 ══════════════════╩════╬═
                                           ║
           lcr_1: 0 ═══════════════════════╩═
        """
        qreg = QuantumRegister(2, "rqr")
        creg = ClassicalRegister(2, "rcr")

        circuit_right = QuantumCircuit(qreg, creg)
        circuit_right.cx(qreg[0], qreg[1])
        circuit_right.tdg(qreg[0])
        circuit_right.measure(qreg, creg)

        # permuted subset of qubits and clbits
        circuit_composed = self.circuit_left.compose(circuit_right, qubits=[1, 4], clbits=[1, 0])

        circuit_expected = self.circuit_left.copy()
        circuit_expected.cx(self.left_qubit1, self.left_qubit4)
        circuit_expected.tdg(self.left_qubit1)
        circuit_expected.measure(self.left_qubit4, self.left_clbit0)
        circuit_expected.measure(self.left_qubit1, self.left_clbit1)

        self.assertEqual(circuit_composed, circuit_expected)

    def test_compose_switch_match(self):
        """Test that composition containing a `switch` with a register that matches proceeds
        correctly."""
        case_0 = QuantumCircuit(1, 2)
        case_0.x(0)
        case_1 = QuantumCircuit(1, 2)
        case_1.z(0)
        case_default = QuantumCircuit(1, 2)
        cr = ClassicalRegister(2, "target")
        right = QuantumCircuit(QuantumRegister(1), cr)
        right.switch(cr, [(0, case_0), (1, case_1), (CASE_DEFAULT, case_default)], [0], [0, 1])

        test = QuantumCircuit(QuantumRegister(3), cr, ClassicalRegister(2)).compose(
            right, [1], [0, 1]
        )

        expected = test.copy_empty_like()
        expected.switch(cr, [(0, case_0), (1, case_1), (CASE_DEFAULT, case_default)], [1], [0, 1])
        self.assertEqual(test, expected)

    def test_compose_switch_no_match(self):
        """Test that composition containing a `switch` with a register that matches proceeds
        correctly."""
        case_0 = QuantumCircuit(1, 2)
        case_0.x(0)
        case_1 = QuantumCircuit(1, 2)
        case_1.z(0)
        case_default = QuantumCircuit(1, 2)
        cr = ClassicalRegister(2, "target")
        right = QuantumCircuit(QuantumRegister(1), cr)
        right.switch(cr, [(0, case_0), (1, case_1), (CASE_DEFAULT, case_default)], [0], [0, 1])
        test = QuantumCircuit(3, 3).compose(right, [1], [0, 1])

        self.assertEqual(len(test.data), 1)
        self.assertIsInstance(test.data[0].operation, SwitchCaseOp)
        target = test.data[0].operation.target
        self.assertIn(target, test.cregs)
        self.assertEqual(list(target), test.clbits[0:2])

    def test_compose_gate(self):
        """Composing with a gate.

                   ┌───┐                               ┌───┐    ┌───┐
        lqr_1_0: ──┤ H ├───                 lqr_1_0: ──┤ H ├────┤ X ├
                   ├───┤                               ├───┤    └─┬─┘
        lqr_1_1: ──┤ X ├───                 lqr_1_1: ──┤ X ├──────┼───
                 ┌─┴───┴──┐     ───■───              ┌─┴───┴──┐   │
        lqr_1_2: ┤ P(0.1) ├  +   ┌─┴─┐   =  lqr_1_2: ┤ P(0.1) ├───┼───
                 └────────┘     ─┤ X ├─              └────────┘   │
        lqr_2_0: ────■─────      └───┘      lqr_2_0: ────■────────┼──
                   ┌─┴─┐                               ┌─┴─┐      │
        lqr_2_1: ──┤ X ├───                 lqr_2_1: ──┤ X ├──────■───
                   └───┘                               └───┘
        lcr_0: 0 ══════════                 lcr_0: 0 ═════════════════

        lcr_1: 0 ══════════                 lcr_1: 0 ═════════════════

        """
        circuit_composed = self.circuit_left.compose(CXGate(), qubits=[4, 0])

        circuit_expected = self.circuit_left.copy()
        circuit_expected.cx(self.left_qubit4, self.left_qubit0)

        self.assertEqual(circuit_composed, circuit_expected)

    def test_compose_one_liner(self):
        """Test building a circuit in one line, for fun."""
        circ = QuantumCircuit(3)
        h = HGate()
        rz = RZGate(0.1)
        cx = CXGate()
        ccx = CCXGate()
        circ = circ.compose(h, [0]).compose(cx, [0, 2]).compose(ccx, [2, 1, 0]).compose(rz, [1])

        expected = QuantumCircuit(3)
        expected.h(0)
        expected.cx(0, 2)
        expected.ccx(2, 1, 0)
        expected.rz(0.1, 1)

        self.assertEqual(circ, expected)

    def test_compose_global_phase(self):
        """Composing with global phase."""
        circ1 = QuantumCircuit(1, global_phase=1)
        circ1.rz(0.5, 0)
        circ2 = QuantumCircuit(1, global_phase=2)
        circ3 = QuantumCircuit(1, global_phase=3)
        circ4 = circ1.compose(circ2).compose(circ3)
        self.assertEqual(
            circ4.global_phase, circ1.global_phase + circ2.global_phase + circ3.global_phase
        )

    def test_compose_front_circuit(self):
        """Test composing a circuit at the front of a circuit."""

        qc = QuantumCircuit(2)
        qc.h(0)
        qc.cx(0, 1)

        other = QuantumCircuit(2)
        other.cz(1, 0)
        other.z(1)

        output = qc.compose(other, front=True)

        expected = QuantumCircuit(2)
        expected.cz(1, 0)
        expected.z(1)
        expected.h(0)
        expected.cx(0, 1)

        self.assertEqual(output, expected)

    def test_compose_front_gate(self):
        """Test composing a gate at the front of a circuit."""

        qc = QuantumCircuit(2)
        qc.h(0)
        qc.cx(0, 1)

        output = qc.compose(CXGate(), [1, 0], front=True)

        expected = QuantumCircuit(2)
        expected.cx(1, 0)
        expected.h(0)
        expected.cx(0, 1)

        self.assertEqual(output, expected)

    def test_compose_adds_parameters(self):
        """Test the composed circuit contains all parameters."""
        a, b = Parameter("a"), Parameter("b")

        qc_a = QuantumCircuit(1)
        qc_a.rx(a, 0)

        qc_b = QuantumCircuit(1)
        qc_b.rx(b, 0)

        with self.subTest("compose with other circuit out-of-place"):
            qc_1 = qc_a.compose(qc_b)
            self.assertEqual(qc_1.parameters, {a, b})

        with self.subTest("compose with other instruction out-of-place"):
            instr_b = qc_b.to_instruction()
            qc_2 = qc_a.compose(instr_b, [0])
            self.assertEqual(qc_2.parameters, {a, b})

        with self.subTest("compose with other circuit in-place"):
            qc_a.compose(qc_b, inplace=True)
            self.assertEqual(qc_a.parameters, {a, b})

    def test_wrapped_compose(self):
        """Test wrapping the circuit upon composition works."""
        qc_a = QuantumCircuit(1)
        qc_a.x(0)

        qc_b = QuantumCircuit(1, name="B")
        qc_b.h(0)

        qc_a.compose(qc_b, wrap=True, inplace=True)

        self.assertDictEqual(qc_a.count_ops(), {"B": 1, "x": 1})
        self.assertDictEqual(qc_a.decompose().count_ops(), {"h": 1, "u3": 1})

    def test_wrapping_unitary_circuit(self):
        """Test a unitary circuit will be wrapped as Gate, else as Instruction."""
        qc_init = QuantumCircuit(1)
        qc_init.x(0)

        qc_unitary = QuantumCircuit(1, name="a")
        qc_unitary.ry(0.23, 0)

        qc_nonunitary = QuantumCircuit(1)
        qc_nonunitary.reset(0)

        with self.subTest("wrapping a unitary circuit"):
            qc = qc_init.compose(qc_unitary, wrap=True)
            self.assertIsInstance(qc.data[1].operation, Gate)

        with self.subTest("wrapping a non-unitary circuit"):
            qc = qc_init.compose(qc_nonunitary, wrap=True)
            self.assertIsInstance(qc.data[1].operation, Instruction)

    def test_condition_mapping_ifelseop(self):
        """Test that the condition in an `IfElseOp` is correctly mapped to a new set of bits and
        registers."""
        base_loose = Clbit()
        base_creg = ClassicalRegister(2)
        base_qreg = QuantumRegister(1)
        base = QuantumCircuit(base_qreg, [base_loose], base_creg)
        with base.if_test((base_loose, True)):
            base.x(0)
        with base.if_test((base_creg, 3)):
            base.x(0)

        test_loose = Clbit()
        test_creg = ClassicalRegister(2)
        test_qreg = QuantumRegister(1)
        test = QuantumCircuit(test_qreg, [test_loose], test_creg).compose(base)

        bit_instruction = test.data[0].operation
        reg_instruction = test.data[1].operation
        self.assertEqual(bit_instruction.condition[0], test_loose)
        self.assertEqual(bit_instruction.condition, (test_loose, True))
        self.assertEqual(reg_instruction.condition[0], test_creg)
        self.assertEqual(reg_instruction.condition, (test_creg, 3))

    def test_condition_mapping_whileloopop(self):
        """Test that the condition in a `WhileLoopOp` is correctly mapped to a new set of bits and
        registers."""
        base_loose = Clbit()
        base_creg = ClassicalRegister(2)
        base_qreg = QuantumRegister(1)
        base = QuantumCircuit(base_qreg, [base_loose], base_creg)
        with base.while_loop((base_loose, True)):
            base.x(0)
        with base.while_loop((base_creg, 3)):
            base.x(0)

        test_loose = Clbit()
        test_creg = ClassicalRegister(2)
        test_qreg = QuantumRegister(1)
        test = QuantumCircuit(test_qreg, [test_loose], test_creg).compose(base)

        bit_instruction = test.data[0].operation
        reg_instruction = test.data[1].operation
        self.assertEqual(bit_instruction.condition[0], test_loose)
        self.assertEqual(bit_instruction.condition, (test_loose, True))
        self.assertEqual(reg_instruction.condition[0], test_creg)
        self.assertEqual(reg_instruction.condition, (test_creg, 3))

    def test_compose_no_clbits_in_one(self):
        """Test combining a circuit with cregs to one without"""
        ansatz = TwoLocal(2, rotation_blocks="ry", entanglement_blocks="cx")

        qc = QuantumCircuit(2)
        qc.measure_all()
        out = ansatz.compose(qc)
        self.assertEqual(out.clbits, qc.clbits)

    def test_compose_no_clbits_in_one_inplace(self):
        """Test combining a circuit with cregs to one without inplace"""
        ansatz = TwoLocal(2, rotation_blocks="ry", entanglement_blocks="cx")

        qc = QuantumCircuit(2)
        qc.measure_all()
        ansatz.compose(qc, inplace=True)
        self.assertEqual(ansatz.clbits, qc.clbits)

    def test_compose_no_clbits_in_one_multireg(self):
        """Test combining a circuit with cregs to one without, multi cregs"""
        ansatz = TwoLocal(2, rotation_blocks="ry", entanglement_blocks="cx")

        qa = QuantumRegister(2, "q")
        ca = ClassicalRegister(2, "a")
        cb = ClassicalRegister(2, "b")
        qc = QuantumCircuit(qa, ca, cb)
        qc.measure(0, cb[1])
        out = ansatz.compose(qc)
        self.assertEqual(out.clbits, qc.clbits)
        self.assertEqual(out.cregs, qc.cregs)

    def test_compose_noclbits_registerless(self):
        """Combining a circuit with cregs to one without, registerless case"""
        inner = QuantumCircuit([Qubit(), Qubit()], [Clbit(), Clbit()])
        inner.measure([0, 1], [0, 1])
        outer = QuantumCircuit(2)
        outer.compose(inner, inplace=True)
        self.assertEqual(outer.clbits, inner.clbits)
        self.assertEqual(outer.cregs, [])

    def test_expr_condition_is_mapped(self):
        """Test that an expression in a condition involving several registers is mapped correctly to
        the destination circuit."""
        inner = QuantumCircuit(1)
        inner.x(0)
        a_src = ClassicalRegister(2, "a_src")
        b_src = ClassicalRegister(2, "b_src")
        c_src = ClassicalRegister(name="c_src", bits=list(a_src) + list(b_src))
        source = QuantumCircuit(QuantumRegister(1), a_src, b_src, c_src)
        target_var = source.add_input("target_var", types.Uint(2))

        test_1 = lambda: expr.lift(a_src[0])
        test_2 = lambda: expr.logic_not(b_src[1])
        test_3 = lambda: expr.logic_and(expr.bit_and(b_src, 2), expr.less(c_src, 7))
        test_4 = lambda: expr.bit_xor(expr.index(target_var, 0), expr.index(target_var, 1))
        source.if_test(test_1(), inner.copy(), [0], [])
        source.if_else(test_2(), inner.copy(), inner.copy(), [0], [])
        source.while_loop(test_3(), inner.copy(), [0], [])
        source.if_test(test_4(), inner.copy(), [0], [])

        a_dest = ClassicalRegister(2, "a_dest")
        b_dest = ClassicalRegister(2, "b_dest")
        dest = QuantumCircuit(QuantumRegister(1), a_dest, b_dest).compose(source)

        # Check that the input conditions weren't mutated.
        for in_condition, instruction in zip((test_1, test_2, test_3), source.data):
            self.assertEqual(in_condition(), instruction.operation.condition)

        # Should be `a_dest`, `b_dest` and an added one to account for `c_src`.
        self.assertEqual(len(dest.cregs), 3)
        mapped_reg = dest.cregs[-1]

        expected = QuantumCircuit(dest.qregs[0], a_dest, b_dest, mapped_reg, inputs=[target_var])
        expected.if_test(expr.lift(a_dest[0]), inner.copy(), [0], [])
        expected.if_else(expr.logic_not(b_dest[1]), inner.copy(), inner.copy(), [0], [])
        expected.while_loop(
            expr.logic_and(expr.bit_and(b_dest, 2), expr.less(mapped_reg, 7)), inner.copy(), [0], []
        )
        # `Var` nodes aren't remapped, but this should be passed through fine.
        expected.if_test(
            expr.bit_xor(expr.index(target_var, 0), expr.index(target_var, 1)),
            inner.copy(),
            [0],
            [],
        )
        self.assertEqual(dest, expected)

    def test_expr_target_is_mapped(self):
        """Test that an expression in a switch statement's target is mapping correctly to the
        destination circuit."""
        inner1 = QuantumCircuit(1)
        inner1.x(0)
        inner2 = QuantumCircuit(1)
        inner2.z(0)

        a_src = ClassicalRegister(2, "a_src")
        b_src = ClassicalRegister(2, "b_src")
        c_src = ClassicalRegister(name="c_src", bits=list(a_src) + list(b_src))
        source = QuantumCircuit(QuantumRegister(1), a_src, b_src, c_src)

        test_1 = lambda: expr.lift(a_src[0])
        test_2 = lambda: expr.logic_not(b_src[1])
        test_3 = lambda: expr.lift(b_src)
        test_4 = lambda: expr.bit_and(c_src, 7)
        source.switch(test_1(), [(False, inner1.copy()), (True, inner2.copy())], [0], [])
        source.switch(test_2(), [(False, inner1.copy()), (True, inner2.copy())], [0], [])
        source.switch(test_3(), [(0, inner1.copy()), (CASE_DEFAULT, inner2.copy())], [0], [])
        source.switch(test_4(), [(0, inner1.copy()), (CASE_DEFAULT, inner2.copy())], [0], [])

        a_dest = ClassicalRegister(2, "a_dest")
        b_dest = ClassicalRegister(2, "b_dest")
        dest = QuantumCircuit(QuantumRegister(1), a_dest, b_dest).compose(source)

        # Check that the input expressions weren't mutated.
        for in_target, instruction in zip((test_1, test_2, test_3, test_4), source.data):
            self.assertEqual(in_target(), instruction.operation.target)

        # Should be `a_dest`, `b_dest` and an added one to account for `c_src`.
        self.assertEqual(len(dest.cregs), 3)
        mapped_reg = dest.cregs[-1]

        expected = QuantumCircuit(dest.qregs[0], a_dest, b_dest, mapped_reg)
        expected.switch(
            expr.lift(a_dest[0]), [(False, inner1.copy()), (True, inner2.copy())], [0], []
        )
        expected.switch(
            expr.logic_not(b_dest[1]), [(False, inner1.copy()), (True, inner2.copy())], [0], []
        )
        expected.switch(
            expr.lift(b_dest), [(0, inner1.copy()), (CASE_DEFAULT, inner2.copy())], [0], []
        )
        expected.switch(
            expr.bit_and(mapped_reg, 7),
            [(0, inner1.copy()), (CASE_DEFAULT, inner2.copy())],
            [0],
            [],
        )

        self.assertEqual(dest, expected)

    def test_join_unrelated_vars(self):
        """Composing disjoint sets of vars should produce an additive output."""
        a = expr.Var.new("a", types.Bool())
        b = expr.Var.new("b", types.Uint(8))

        base = QuantumCircuit(inputs=[a])
        other = QuantumCircuit(inputs=[b])
        out = base.compose(other)
        self.assertEqual({a, b}, set(out.iter_vars()))
        self.assertEqual({a, b}, set(out.iter_input_vars()))
        # Assert that base was unaltered.
        self.assertEqual({a}, set(base.iter_vars()))

        base = QuantumCircuit(captures=[a])
        other = QuantumCircuit(captures=[b])
        out = base.compose(other)
        self.assertEqual({a, b}, set(out.iter_vars()))
        self.assertEqual({a, b}, set(out.iter_captured_vars()))
        self.assertEqual({a}, set(base.iter_vars()))

        base = QuantumCircuit(inputs=[a])
        other = QuantumCircuit(declarations=[(b, 255)])
        out = base.compose(other)
        self.assertEqual({a, b}, set(out.iter_vars()))
        self.assertEqual({a}, set(out.iter_input_vars()))
        self.assertEqual({b}, set(out.iter_declared_vars()))

    def test_var_remap_to_avoid_collisions(self):
        """We can use `var_remap` to avoid a variable collision."""
        a1 = expr.Var.new("a", types.Bool())
        a2 = expr.Var.new("a", types.Bool())
        b = expr.Var.new("b", types.Bool())

        base = QuantumCircuit(inputs=[a1])
        other = QuantumCircuit(inputs=[a2])

        out = base.compose(other, var_remap={a2: b})
        self.assertEqual([a1, b], list(out.iter_input_vars()))
        self.assertEqual([a1, b], list(out.iter_vars()))

        out = base.compose(other, var_remap={"a": b})
        self.assertEqual([a1, b], list(out.iter_input_vars()))
        self.assertEqual([a1, b], list(out.iter_vars()))

        out = base.compose(other, var_remap={"a": "c"})
        self.assertTrue(out.has_var("c"))
        c = out.get_var("c")
        self.assertEqual(c.name, "c")
        self.assertEqual([a1, c], list(out.iter_input_vars()))
        self.assertEqual([a1, c], list(out.iter_vars()))

    def test_stretch_remap_to_avoid_collisions(self):
        """We can use `var_remap` to avoid a stretch collision."""
        a1 = expr.Stretch.new("a")
        a2 = expr.Stretch.new("a")
        b = expr.Stretch.new("b")

        base = QuantumCircuit(captures=[a1])
        other = QuantumCircuit(captures=[a2])

        out = base.compose(other, var_remap={a2: b})
        self.assertEqual([a1, b], list(out.iter_captured_stretches()))
        self.assertEqual([a1, b], list(out.iter_stretches()))

        out = base.compose(other, var_remap={"a": b})
        self.assertEqual([a1, b], list(out.iter_captured_stretches()))
        self.assertEqual([a1, b], list(out.iter_stretches()))

        out = base.compose(other, var_remap={"a": "c"})
        self.assertTrue(out.has_stretch("c"))
        c = out.get_stretch("c")
        self.assertEqual(c.name, "c")
        self.assertEqual([a1, c], list(out.iter_captured_stretches()))
        self.assertEqual([a1, c], list(out.iter_stretches()))

    def test_simple_inline_captures(self):
        """We should be able to inline captures onto other variables."""
        a = expr.Var.new("a", types.Bool())
        b = expr.Var.new("b", types.Bool())
        c = expr.Var.new("c", types.Uint(8))
        d = expr.Stretch.new("d")

        base = QuantumCircuit(inputs=[a, b])
        base.add_var(c, 255)
        base.add_stretch(d)
        base.store(a, expr.logic_or(a, b))
        other = QuantumCircuit(captures=[a, b, c, d])
        other.store(c, 254)
        other.store(b, expr.logic_or(a, b))
        new = base.compose(other, inline_captures=True)

        expected = QuantumCircuit(inputs=[a, b])
        expected.add_var(c, 255)
        expected.add_stretch(d)
        expected.store(a, expr.logic_or(a, b))
        expected.store(c, 254)
        expected.store(b, expr.logic_or(a, b))
        self.assertEqual(new, expected)

    def test_can_inline_a_capture_after_remapping(self):
        """We can use `var_remap` to redefine a capture variable _and then_ inline it in deeply
        nested scopes.  This is a stress test of capture inlining."""
        a = expr.Var.new("a", types.Bool())
        b = expr.Var.new("b", types.Bool())
        c = expr.Var.new("c", types.Uint(8))

        # We shouldn't be able to inline `qc`'s variable use as-is because it closes over the wrong
        # variable, but it should work after variable remapping.  (This isn't expected to be super
        # useful, it's just a consequence of how the order between `var_remap` and `inline_captures`
        # is defined).
        base = QuantumCircuit(inputs=[a])
        qc = QuantumCircuit(declarations=[(c, 255)], captures=[b])
        qc.store(b, expr.logic_and(b, b))
        with qc.if_test(expr.logic_not(b)):
            with qc.while_loop(b):
                qc.store(b, expr.logic_not(b))
            # Note that 'c' is captured in this scope, so this is also a test that 'inline_captures'
            # doesn't do something silly in nested scopes.
            with qc.switch(c) as case:
                with case(0):
                    qc.store(c, expr.bit_and(c, 255))
                with case(case.DEFAULT):
                    qc.store(b, expr.equal(c, 255))
        base.compose(qc, inplace=True, inline_captures=True, var_remap={b: a})

        expected = QuantumCircuit(inputs=[a], declarations=[(c, 255)])
        expected.store(a, expr.logic_and(a, a))
        with expected.if_test(expr.logic_not(a)):
            with expected.while_loop(a):
                expected.store(a, expr.logic_not(a))
            # Note that 'c' is not remapped.
            with expected.switch(c) as case:
                with case(0):
                    expected.store(c, expr.bit_and(c, 255))
                with case(case.DEFAULT):
                    expected.store(a, expr.equal(c, 255))

        self.assertEqual(base, expected)

    def test_rejects_duplicate_bits(self):
        """Test that compose rejects duplicates in either qubits or clbits."""
        base = QuantumCircuit(5, 5)

        attempt = QuantumCircuit(2, 2)
        with self.assertRaisesRegex(CircuitError, "Duplicate qubits"):
            base.compose(attempt, [1, 1], [0, 1])
        with self.assertRaisesRegex(CircuitError, "Duplicate clbits"):
            base.compose(attempt, [0, 1], [1, 1])

    def test_cannot_mix_inputs_and_captures(self):
        """The rules about mixing `input` and `capture` vars should still apply."""
        a = expr.Var.new("a", types.Bool())
        b = expr.Var.new("b", types.Uint(8))
        c = expr.Stretch.new("c")
        with self.assertRaisesRegex(CircuitError, "circuits with input variables cannot be"):
            QuantumCircuit(inputs=[a]).compose(QuantumCircuit(captures=[b]))
        with self.assertRaisesRegex(CircuitError, "circuits to be enclosed with captures cannot"):
            QuantumCircuit(captures=[a]).compose(QuantumCircuit(inputs=[b]))
        with self.assertRaisesRegex(CircuitError, "circuits with input variables cannot be"):
            QuantumCircuit(inputs=[a]).compose(QuantumCircuit(captures=[c]))

    def test_reject_var_naming_collision(self):
        """We can't have multiple vars with the same name."""
        a1 = expr.Var.new("a", types.Bool())
        a2 = expr.Var.new("a", types.Bool())
        b = expr.Var.new("b", types.Bool())
        self.assertNotEqual(a1, a2)

        s1 = expr.Stretch.new("s")
        s2 = expr.Stretch.new("s")
        t = expr.Stretch.new("t")
        self.assertNotEqual(s1, s2)

        with self.assertRaisesRegex(CircuitError, "cannot add.*shadows"):
            QuantumCircuit(inputs=[a1]).compose(QuantumCircuit(inputs=[a2]))
        with self.assertRaisesRegex(CircuitError, "cannot add.*shadows"):
            QuantumCircuit(captures=[s1]).compose(QuantumCircuit(captures=[s2]))
        with self.assertRaisesRegex(CircuitError, "cannot add.*shadows"):
            QuantumCircuit(captures=[a1]).compose(QuantumCircuit(declarations=[(a2, False)]))
        with self.assertRaisesRegex(CircuitError, "cannot add.*shadows"):
            QuantumCircuit(declarations=[(a1, True)]).compose(
                QuantumCircuit(inputs=[b]), var_remap={b: a2}
            )
        with self.assertRaisesRegex(CircuitError, "cannot add.*shadows"):
            QuantumCircuit(captures=[s1]).compose(QuantumCircuit(captures=[t]), var_remap={t: s2})

    def test_reject_remap_var_to_bad_type(self):
        """Can't map a var to a different type."""
        a = expr.Var.new("a", types.Bool())
        b = expr.Var.new("b", types.Uint(8))
        qc = QuantumCircuit(inputs=[a])
        with self.assertRaisesRegex(CircuitError, "mismatched types"):
            QuantumCircuit().compose(qc, var_remap={a: b})
        qc = QuantumCircuit(captures=[b])
        with self.assertRaisesRegex(CircuitError, "mismatched types"):
            QuantumCircuit().compose(qc, var_remap={b: a})

    def test_reject_remap_identifier_to_wrong_kind(self):
        """Can't map a var to stretch or stretch to var."""
        a = expr.Var.new("a", types.Bool())
        b = expr.Stretch.new("b")
        qc = QuantumCircuit(inputs=[a])
        with self.assertRaisesRegex(CircuitError, "mismatched identifier"):
            QuantumCircuit().compose(qc, var_remap={a: b})
        qc = QuantumCircuit(captures=[b])
        with self.assertRaisesRegex(CircuitError, "mismatched identifier"):
            QuantumCircuit().compose(qc, var_remap={b: a})

    def test_reject_inlining_missing_var(self):
        """Can't inline a var that doesn't exist."""
        a = expr.Var.new("a", types.Bool())
        b = expr.Var.new("b", types.Bool())
        qc = QuantumCircuit(captures=[a])
        with self.assertRaisesRegex(CircuitError, "Variable '.*' to be inlined is not in the base"):
            QuantumCircuit().compose(qc, inline_captures=True)

        # 'a' _would_ be present, except we also say to remap it before attempting the inline.
        qc = QuantumCircuit(captures=[a])
        with self.assertRaisesRegex(CircuitError, "Replacement '.*' for variable '.*' is not in"):
            QuantumCircuit(inputs=[a]).compose(qc, var_remap={a: b}, inline_captures=True)

    def test_reject_inlining_missing_stretch(self):
        """Can't inline a var that doesn't exist."""
        a = expr.Stretch.new("a")
        b = expr.Stretch.new("b")
        qc = QuantumCircuit(captures=[a])
        with self.assertRaisesRegex(CircuitError, "Variable '.*' to be inlined is not in the base"):
            QuantumCircuit().compose(qc, inline_captures=True)

        # 'a' _would_ be present, except we also say to remap it before attempting the inline.
        qc = QuantumCircuit(captures=[a])
        with self.assertRaisesRegex(CircuitError, "Replacement '.*' for variable '.*' is not in"):
            QuantumCircuit().compose(qc, var_remap={a: b}, inline_captures=True)


if __name__ == "__main__":
    unittest.main()
